{  SILAB PPC Programmer.spin

          **************************************
           Flash Silabs micros via C2 interface
          **************************************
                    Peter Jakacki 2009                 
          **************************************

PPC chips based upon Silabs C8051F3xx can be programmed over
the C2 interface using 2 wires. In a Propeller system one of the
C2 signals is shared with the I2C bus SDA while the other is a
dedicated Reset/Clock (C2CLK) combo.

The C2CLK could be shared with another pin via a jumper if resources
are scarce.

This is a run-once program which once it has programmed the PPC no
longer needs to be resident so use F10 from the Spin tool or
run from RAM.

}
''****************************************************************    

CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000
  clockfreq = ((_CLKMODE - XTAL1) >> 6) * _XINFREQ

   _1mS  = clockfreq /     1_000 'Divisor for  1 mS


  C2CK2         = 19
  C2CK1         = 23
  C2CK0         = 25
  C2D           = 29            ' uses same pin as SDA on Prop
  
' coms  
  txdcoms        = 30
  rxdcoms        = 31 


' C2 constants
  DEVID         = 0
  FPCTL         = 2
  FPDAT         = $B4
  inbusy        = 2
  outready      = 1

  
'
VAR
  long  faddr,flen,flen2,cold
  word  c2cmd,c2dat
  
      
DAT

VER       byte " HARDWARE: PROPELLER",$0d,$0a
          byte " FIRMWARE: C2 Flash Loader V1.0",$0d,$0a,0

PPC
        long  @endPPC-@PPC
        file  "PPC.BIN"
endPPC

CHIPS
        byte  $F2,$F4,$F0       ' list of I2C address for each chip
endCHIPS


PUB Main | i
  if cold == $a55a               ' only want to run once - in case this has been loaded into EEPROM
    REBOOT
  cognew(@C2CODE,@c2cmd)
  i~
  repeat endCHIPS-CHIPS
     C2SET(C2CK0)
     byte[@PPC][$84] := byte[@CHIPS][i++]                ' Set I2C address of chip
     C2RESET
     \PROGRAM(@PPC)
  cold := $A55A      
  REBOOT
      
  
pub PROGRAM(fileptr)
  C2RESET
  if RDATA <> $0A                                       ' C8051F33x?
    abort string(" NO DEVICE")
  C2RESET
  C2INIT
  C2DeviceErase                        
  VerifyErased
  faddr := 0
  flen := long[fileptr]
  fileptr += 4
  repeat (flen>>8)+1
    flen2 := flen
    flen2 <#= 256
    ProgramBlock(fileptr,faddr,flen2)
    faddr += 256
    flen -= 256
  C2RESET

pub ProgramBlock(src,addr,len) | i
  WADDRESS(FPDAT)
  WDATA(07)
  if RDATAW <> $0d                                                                                      
    abort string(" REJECT COMMAND")
  WDATA(addr>>8)                      ' ADDRH
  WDATA(addr&$ff)                      ' ADDRL
  WDATA(len)
  if RDATAW <> $0d
    abort string(" REJECT ADDRESS")
  repeat i from addr to addr+len-1
    WDATA(byte[i][src])  

pub PollInBusy
  repeat while RSTATUS&inbusy
  
pub PollOutReady
  repeat until RSTATUS&outready

pub VerifyErased

pub C2DeviceErase
  WADDRESS(FPDAT)
  WDATA(03)
  if RDATAW <> $0d                                                                                      
    abort string(" REJECT ERASE,")
  WDATA($DE)
  WDATA($AD)
  WDATA($A5)
  if RDATAW <> $0d                                                                                      
    abort string(" ERASE FAIL,")

pub command(cmd,data)
  c2dat := data
  c2cmd := cmd
  repeat until c2cmd == 0
  result := c2dat
  
pub C2RESET
  command("Z",0)                  ' reset

pub C2SET(data)
  command("=",data)


pub RDATA
  result := command("@",0)

pub RDATAW
  PollOutReady
  result := RDATA

pub WDATA(data)
  command("!",data)
  PollInBusy

pub WDATA1(data)
  command("!",data)


pub WADDRESS(data)
  command(":",data)

pub RSTATUS
  result := command("?",0)

pub C2INIT
  WADDRESS(FPCTL)
  WDATA(2)
  WDATA(1)
  ms(20)

PUB ms(Period)                                    '1 mS 
    waitcnt(_1mS * Period + cnt)


DAT
              org   0
C2CODE
              mov       par2,par
              add       par2,#2
              or        outa,_c2d
              or        dira,_c2d
CLEAR
              wrword    zero,par2
C2CMDS        wrword    zero,par                  ' clear command
C2CMDLP
              rdword    r0,par                  ' read command
              cmp       r0,#0 wz
        if_z  jmp       #C2CMDLP
              cmp       r0,#"Z"  wz             ' Reset command?
        if_z  jmp       #Reset
              cmp       r0,#"@" wz              ' data ?
        if_z  jmp       #ReadData
              cmp       r0,#":" wz              ' Write Address?
        if_z  jmp       #WriteAddress
              cmp       r0,#"!" wz
        if_z  jmp       #WriteData
              cmp       r0,#"?" wz              ' read status
        if_z  jmp       #ReadAddress
              cmp       r0,#"=" wz              ' set c2ck
        if_z  jmp       #SetC2
              jmp       #C2CMDS

SetC2
              rdword    r0,par2
              mov       _c2ck,#1
              shl       _c2ck,r0
              jmp       #C2CMDS
                             
WriteAddress
              rdword    r0,par2    
              call      #WA
              jmp       #C2CMDS

ReadAddress
              call      #RA
              jmp       #WRPAR 

WriteData
              rdword    r0,par2    
              call      #WD
              jmp       #C2CMDS
              
ReadData
              call      #RD
WRPAR
              wrword    r0,par2    
              jmp       #C2CMDS 
              

'************************* C2 PRIMITIVES *********************        

Reset         call      #RS
              jmp       #C2CMDS

RS
              andn      outa,_c2ck
              or        dira,_c2ck
              mov       r0,#250
rstdly        nop
              djnz      r0,#rstdly
              or        outa,_c2ck
RS_ret        ret

INS
   '' INS field (00b, LSB first)
              shr       r2,#1 wc
              muxc      outa,_c2d
              or        dira,_c2d               '' Enable C2D driver (output)
              nop
              andn      outa,_c2ck              ''strobe 
              or        outa,_c2ck
              nop
              shr       r2,#1 wc
              muxc      outa,_c2d
              nop
              andn      outa,_c2ck              ''strobe 
              or        outa,_c2ck
INS_ret       ret

WAITRDY
              andn      dira,_c2d               '' Disable C2D driver for input
              andn      outa,_c2ck              ''strobe
              or        outa,_c2ck
              test      _c2d,ina wz
        if_z  jmp       #Waitrdy                 '' Strobe C2CK until target transmits a '1' 
WAITRDY_ret   ret


READ8
              andn      dira,_c2d               '' C2D disabled
              mov       r0,#0
              mov       r1,#8
ReadDRlp
              shr       r0,#1
              andn      outa,_c2ck              ''strobe 
              or        outa,_c2ck
              test      _c2d,ina wz
              muxnz     r0,#$80
              djnz      r1,#ReadDRlp
READ8_ret     ret

WRITE8
              or        dira,_c2d
              mov       r1,#8
WRDAT
              shr       r0,#1 wc
              muxc      outa,_c2d      
              andn      outa,_c2ck              ''strobe
              or        outa,_c2ck
              djnz      r1,#WRDAT
WRITE8_ret    ret

start
              andn      dira,_c2d
              andn      outa,_c2ck              '' START - Strobe C2CK with C2D driver disabled
              or        outa,_c2ck
start_ret     ret

stop
              or        outa,_c2d               '' pull it high just for diagnostics
              andn      dira,_c2d               '' STOP
              andn      outa,_c2ck              ''strobe
              or        outa,_c2ck
stop_ret      ret        

'*************************** READ DATA REGISTER ***************************

RD
              call      #start
              mov       r2,#0
              call      #INS
              mov       r2,#0
              call      #INS
              call      #WAITRDY
              call      #READ8
              call      #STOP
RD_ret        ret

'************************* READ ADDRESS REGISTER **************************

RA
              call      #start
              mov       r2,#02
              call      #INS
              call      #Read8
              call      #STOP
RA_ret        ret

'************************** WRITE DATA REGISTER ***************************
              
WD
              call      #start
              mov       r2,#1
              call      #INS
              mov       r2,#0
              call      #INS
              call      #WRITE8
              call      #WAITRDY
              call      #stop
WD_ret        ret

'****************************** WRITE ADDRESS REGISTER *************************
              
WA            call      #start
              mov       r2,#3
              call      #INS
              call      #WRITE8
              call      #stop
WA_ret        ret


_c2ck   long  0
_c2d    long  |<C2D
zero    long  0
_ack    long  $8000_0000


r0      res   1
r1      res   1
r2      res   1
par2    res   1

DAT
{
PPC - Analog and Parallel I/O

Many times I have needed dedicated analog inputs for the Propeller with at least 10-bit resolution.
Although there are plenty of A/D converters available that would be suitable it is a different story when it comes price.
In other projects it has been easier and cheaper for me to use a humble microcontroller with on-board A/D.
The same goes for the Propeller when I need DC accurate A/D.

What I have done is implemented an auto-scanning A/D plus parallel I/O expansion in a cheap micro that can
interface to the Prop in a variety of ways. I2C is the common method as is SPI plus also there are the serial UART modes.
At present I have this implemented on a Silabs C8051F334 chip which a small 20 pin QFN about the size of a 3mm LED
but I am also looking at implementing this on similar MC9S08 chips that are available in a variety of packages
including DIP.

When I say cheap I mean they are available for a buck of two from Mouser.

Of course this means nothing if you have to write your own software and find a way to program them.
That's why I have already sorted this matter out (after much hair loss) and the chips can be connected
to a Propeller and loaded from a small Spin loader.
At present I have written firmware for the Silabs part that surpasses my expectations and the good thing
is that it is completely customizable to suit the application.

When I want to upload new firmware I simply select the C2 loader in the Spin tool and hit F10.
The loader runs automatically of course and the firmware files which are embedded in the Spin program
via the FILE directive are loaded into the PPC within a couple of seconds.
After this the loader reboots and you are back to where you were before the interruption as the Prop loads
it's default program from the EEPROM.

I have one system where I have 2 of these chips acting as A/D converters and parallel I/O plus another chip
that runs a front panel with switch inputs and about 40 string of LEDs.
All these chips are connected up to the same I2C lines as the Prop's EEPROM but I have dedicated
an I/O line per chip just for in-circuit reprogramming from the Prop.



PROP PPC (based on C8051F33x)

Interfaces: I2C, SPI, UART
Analog inputs: up to 14 channels x10-bit + temperature
Parallel: up to 14 input/output/quasi  5V tolerant
Peripherals: UART,SPI,I2C,4 Counter/timers

The Silabs parts run at 25MIPS and include a calibrated on-board oscillator +/-2%

Additionally the MC9S08SH4 version includes a low-power RTC.

To talk to my demo PPC chip.
ADDRESS PCC on I2C
  READING:
    READ 10 bytes as STATUS[1],PORT1[1],ADO[2],AD1[2],AD2[2],AD3[2]
  WRITING:
    COMMAND[1],<DATA>

MY demo PPC COMMANDS:
$00 NOP
$01 WRITE PORT 1 using supplied data and <direction>
    i.e. $01,$F0,$FC writes F0 to port 1 with outputs on 2..7
$02 WRITE PORT 1 DIR REGISTER using supplied data
$FD READ VERSION STRING for up to 32 bytes
$FE READ PROGRAM MEMORY for 32 bytes using supplied address (ADRL,ADRH)
    i.e. $11,$A0,$00 will read the the extended part of the version string at $00A0
$FF RESET PPC


}
